#!/usr/bin/env python
# encoding: utf-8
"""
Author: Nick Bearson, nickb@ssec.wisc.edu
Copyright (c) 2012 University of Wisconsin SSEC. All rights reserved.
"""

from sensor import Sensor
from descriptors import reify

import numpy as np

from pyhdf import SD

from datetime import datetime, timedelta

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

#ctp_path = '/odyssey/trends/verify/HIRS/ctp/orbit/stats/noaa-18/2009/' # TODO: hardcoded path very bad
ctp_path = '/odyssey/trends/ncdc/ctp/orbit/stats/noaa-18/2009/' # mjh 3/18/2015 fixed path for Nick's reprocessed CTP files
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# FIXME: FUNCTIONS THAT MAY OR MAY NOT BELONG HERE, SHOULD THEY MOVE TO SENSOR OR A UTIL MODULE?

def _search_list_regex(alist, regex):
  """
  Utility function to search a list of strings and return a list of any that match the given regular expression.
  """
  import re # FIXME: move this function, and move this import when you do
  retlist = []
  for i in alist:
    if re.match(regex, i):
      retlist.append(i)
  return retlist

ref_bands = ['1', '2', '3a']
emis_bands = ['3b', '4', '5']

avhrr_band_lookup = {   # (rsb, emis, bt)
  '1'  : ('refl_0_65um_nom', None, None),
  '2'  : ('refl_0_86um_nom', None, None),
  '3a' : ('refl_1_60um_nom', None, None),
  # mjh: 3.75, 11, and 12 should all be in both EmissiveBands and BrightnessTemperatureBands
  '3b' : (None, 'temp_3_75um_nom', 'temp_3_75um_nom'), # FIXME: wtf?  refl w/ bt?  is this a bug with CLAVRx?
  '4'  : (None, 'temp_11_0um_nom', 'temp_11_0um_nom'),
  '5'  : (None, 'temp_12_0um_nom', 'temp_12_0um_nom'),
}

avhrr_band_centers = {
  '1' : 0.65,
  '2' : 0.86,
  '3a': 1.60,
  '3b': 3.75,
  '4' : 11.0,
  '5' : 12.0,
}

#hirs_bands = [4, 5, 6, 7, 8, 9, 10, 11, 12, 19]
hirs_bands = [4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19]

#hirs_centers is in rad_to_BT_constants[0]
hirs_channel_map = { 1  : "14.95",
                2  : "14.71",
                3  : "14.49",
                4  : "14.22",
                5  : "13.97",
                6  : "13.64",
                7  : "13.35",
                8  : "11.11",
                9  : "9.71",
                10 : "12.47",
                11 : "7.33",
                12 : "6.52",
                13 : "4.57",
                14 : "4.52",
                15 : "4.47",
                16 : "4.45",
                17 : "4.13",
                18 : "4.00",
                19 : "3.76",
                20 : "690000" }

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


def _scale_and_offset(sdsname, fname):
  """
  Generic function to read an sds from our file & return it with scale & offset applied
  """
  sd = SD.SD(fname)
  r = sd.select(sdsname)
  data = r[:].astype(np.float32)
  # mjh deal with fill values properly
  fill_value = getattr(r, '_FillValue', None)
  if fill_value:
      fill_index = data == r._FillValue
  try:
    data = data * r.scale_factor
  except:
    pass
  try:
    data = data + r.add_offset
  except:
    pass
  # mjh scale reflectances from 0 to 1 instead of percent
  if sdsname.startswith('refl'):
      data = data / 100.0
  # mjh deal with fill values properly
  if fill_value:
      data[fill_index] = -999.0
  return data

class AVHRR(Sensor):
  """
  AVHRR sensor object to give the IFF implementation everything it needs from l2 patmosx files
  """
  SENSOR_EPOCH = datetime(1958, 1, 1)
  GROUPING_REGEX = '.*NSS\.GHRR\.[A-Z]{2}\.D([0-9]{2})([0-9]{3})\.S([0-9]{4})\..*\.hdf'
  nDETECTORS = 1 # used for doing scan line time math - we have a time for each line though

  def __init__(self, tag, platform, *args, **kwargs):
    # At the moment we have no extra args, but process them like this if we do...
    try:
      pass
# FIXME: CURRENTLY A NULL OP
#      self._w = kwargs.pop('w')
    except KeyError:
      pass

    super(AVHRR, self).__init__(*args, **kwargs)

    log.debug(self._file_list)

    HIRS_REGEX = '.*/?NSS\.HIRX'

    self.SENSOR = "avhrr"
    self.PLATFORM = platform
    self.tag = tag

    self._geo_file = _search_list_regex(self._file_list, AVHRR.GROUPING_REGEX)[0]
    self._img_file = self._geo_file # geo & img are in the same file for avhrr
    self._hirs_files = sorted(_search_list_regex(self._file_list, HIRS_REGEX))

    self._refs = None
    self._ref_wav = None
    self._ref_nums = None

    self._emis = None
    self._emis_wav = None
    self._emis_nums = None


  @staticmethod
  def date_bounds_from_fname(fname):
    DATE_REGEX = ".*D([0-9]{2})([0-9]{3})\.S([0-9]{4})\.E([0-9]{4})\..*.hdf"
    import re
    m = re.match(DATE_REGEX, fname)
    if m:
      yy, ddd, ssss, eeee = m.group(1), m.group(2), m.group(3), m.group(4)
      frmt = '%y%j%H%M'
      start = datetime.strptime(yy+ddd+ssss, frmt)
      end = datetime.strptime(yy+ddd+eeee, frmt)
      if start > end:
          end = end + timedelta(days=1)
      return start, end        
    else:
      raise RuntimeError, "can't get valid date bounds from file:\n%s\ndoesn't fit format:\n%s" % (fname, DATE_REGEX)

  @reify
  def ReflectiveSolarBands(self):
    l = []
    for b in ref_bands:
      sdsname = avhrr_band_lookup[b][0]
      if sdsname is not None:
        d = _scale_and_offset(sdsname, self._img_file)
      else:
        d = (self.Latitude * 0) - 999.0  # Fill
      l.append(d)
    return np.array(l)

  @reify
  def ReflectiveBandCenters(self):
    # need to specify type here because it doesn't get done in aggregator.py for this case.
    return np.array([avhrr_band_centers[b] for b in ref_bands]).astype(np.float32)

  @reify
  def ReflectiveBandNums(self):
    return np.array(ref_bands)
    
  @reify
  def EmissiveBands(self):
    # mjh import custom code for simple planck function evaluation
    from planck import planck_wl
    l = []
    for b in emis_bands:
      sdsname = avhrr_band_lookup[b][1]
      if sdsname is not None:
        d = _scale_and_offset(sdsname, self._img_file)

        # mjh convert brightness temperatures to radiances.
        # would be nice to guarantee we are only touching AVHRR bands here,
        # but for now we can get away with not checking.
        log.debug('Converting AVHRR Tb\'s to rads; channel %s @ %f um' % (b, avhrr_band_centers[b]))
        d[d != -999.0] = planck_wl(avhrr_band_centers[b], d[d != -999.0])

      else:
        d = (self.Latitude * 0) - 999.0  # Fill
      l.append(d)
    return np.array(l)

  @reify
  def EmissiveBandCenters(self):
    return np.array([avhrr_band_centers[b] for b in emis_bands])

  @reify
  def EmissiveBandNums(self):
    return np.array(emis_bands)

  @reify
  def BrightnessTemperatureBands(self):
    l = []
    for b in emis_bands:
      sdsname = avhrr_band_lookup[b][2]
      if sdsname is not None:
        d = _scale_and_offset(sdsname, self._img_file)
      else:
        d = (self.Latitude * 0) - 999.0  # Fill
      l.append(d)
    return np.array(l)
    
  @reify
  def BrightnessTemperatureBandCenters(self):
    return np.array([avhrr_band_centers[b] for b in emis_bands])

  @reify
  def BrightnessTemperatureBandNums(self):
    return np.array(emis_bands)

  @reify
  def _HirsMapping(self):
    import h2a
    assert (len(self._hirs_files) == 1) # mjh TODO: generalize to n hirs files
    return h2a.hirs2avhrr_transform(self._hirs_files[0], self._geo_file)

  @reify
  def HirsBands(self):
    import h2a
    log.debug('making hirs bands!')

    hirsshape = (len(hirs_bands),) + self.Latitude.shape
    allbts  = np.empty(hirsshape)
    allbts.fill(-999.0)
    allrads = np.empty(hirsshape)
    allrads.fill(-999.0)

    for i in range(len(self._hirs_files)):
      hirs = self._hirs_files[i]
      avhrr_geo = self._geo_file

      log.debug("hirsBands processing...")
      log.debug("hirs file: %s" % (hirs))
      log.debug("avhrr geo file: %s" % (avhrr_geo))

      #mapping = h2a.hirs2avhrr_transform(hirs, avhrr_geo)
      mapping = self._HirsMapping

      radlist = []
      btlist = []
      centerlist = []

      for band in hirs_bands:
        log.debug("adding hirs band %d" % (band))
        brads = _scale_and_offset('channel_' + str(band) + '_radiance', hirs)
        bbts = _scale_and_offset('channel_' + str(band) + '_brightness_temperature', hirs)
        vrads = h2a.apply_mapping(mapping, brads)
        vbts = h2a.apply_mapping(mapping, bbts)

        # mjh need to convert "NOAA radiances" [mW/cm^2/cm^-1/str] to "NASA radiances" [W m-2 um-1 sr-1]
        # following modis_module.f90 in clavr-x and noting clavr-x wavenumbers are in [1/cm], our wavelengths are in [um]
        vrads[vrads != -999.0] = (10.0*vrads[vrads != -999.0]) / (float(hirs_channel_map[band])**2.0)

        radlist.append(vrads)
        btlist.append(vbts)

      rads = np.array(radlist, dtype=radlist[0].dtype)
      bts = np.array(btlist, dtype=btlist[0].dtype)
      centers = np.array(centerlist)

      # mix the new array into the total array - FIXME: REMEMBER WE'RE USING NEAREST - ADD WEIGHTING WHEN ITS AVAILABLE
      def combinate(total, new):
        if total == None:
          log.debug("combining hirs granules: first in set")
          total = new
        else:
          assert( total.shape == new.shape ) # new should always be the same shape as total... right? catch if it's not because that could break other things
          total[total == -999.0] = new[total == -999.0] # FIXME: -999.0 is just our fill - replace this with checking for larger weights?
        return total

      log.debug("allrads: %s allbts: %s rads: %s bts: %s" % (allrads.shape, allbts.shape, rads.shape, bts.shape))
      allrads = combinate(allrads, rads)
      allbts = combinate(allbts, bts)

    log.debug("returning from hirsBands")
    # mjh: final result shouldn't have nan's (esp. scary cause they will come through as valid data values after
    # clavr-x processing... e.g. 260K for BT's
    allrads[np.isnan(allrads)] = -999.0
    allbts[np.isnan(allbts)] = -999.0
    return allrads, allbts

  @reify
  def SounderEmissive(self):
    return self.HirsBands[0]

  @reify
  def SounderBTs(self):
    return self.HirsBands[1]

  @reify
  def SounderCenters(self):
    return np.array([float(hirs_channel_map[band]) for band in hirs_bands])
    #return np.array(hirs_bands) # FIXME: fix this to actually grab the numbers, just testing for now though

  @reify
  def SounderNums(self):
    return np.array(['HIRS-'+str(band) for band in hirs_bands])
    #return np.array(hirs_bands)

  @reify
  def Height(self):
   return (self.Latitude * 0) - 999  # FIXME: Fill for now

  @reify
  def Range(self):
   return (self.Latitude * 0) - 999  # FIXME: Fill for now

  @reify
  def SensorZenith(self):
    return _scale_and_offset('sensor_zenith_angle', self._img_file)

  @reify
  def SensorAzimuth(self):
    return _scale_and_offset('relative_azimuth_angle', self._img_file)
    #return (self.Latitude * 0) - 999  # FIXME: Fill for now

  @reify
  def SolarZenith(self):
    return _scale_and_offset('solar_zenith_angle', self._img_file)

  @reify
  def SolarAzimuth(self):
    return _scale_and_offset('solar_azimuth_angle', self._img_file)

  @reify
  def Latitude(self):
    return _scale_and_offset('latitude', self._img_file)

  @reify
  def Longitude(self):
    return _scale_and_offset('longitude', self._img_file)

  @reify
  def ScanStartTime(self):
    slt = _scale_and_offset('scan_line_time', self._img_file) * 3600 # scan line time in seconds
    start, end = self.date_bounds_from_fname(self._geo_file)
    daystart = datetime(start.year, start.month, start.day)
    day_seconds = (daystart - self.SENSOR_EPOCH).days * 86400 # daystart will always = days with no seconds
    slt = slt + day_seconds     # FIXME: need to handle day swap!
    return slt

  @reify
  def SounderElementIndex(self):
    mapping = self._HirsMapping[:,:,1]
    return mapping.filled(fill_value=-999).astype(np.int32)

  @reify
  def SounderLineIndex(self):
    mapping = self._HirsMapping[:,:,0]
    return mapping.filled(fill_value=-999).astype(np.int32)

  @reify
  def SounderCloudHeights(self):
    import string, h2a
    from netCDF4 import Dataset
    from os.path import basename, isfile
    assert (len(self._hirs_files) == 1) # mjh TODO: generalize to n hirs files
    hirs = self._hirs_files[0]
    cth_file = string.replace(basename(hirs), '.hdf', '')+'.ctp.nc' # mjh 3/18/2015 added .ctp
    if isfile(ctp_path+cth_file): # mjh 3/18/2015 fixed bug (path wasn't included)
        log.debug('cth_file found: %s' % cth_file)
    else:
        log.debug('cth_file NOT found, filling with junk: %s' % cth_file)
        return ((self.Latitude * 0.0) - 999.0).astype(np.float32)
    rootgrp = Dataset(ctp_path+cth_file, 'r')
    cth_data = np.array(rootgrp.variables['cloud_top_height'])
    ret = h2a.apply_mapping(self._HirsMapping, cth_data).astype(np.float32)
    ret[ret < 0.0] = -999.0  # weed out -99.9's ... not sure where the -99.9's come from???
    #TODO(mjh):  are the zeroes legitimate?
    #rootgrp.close()  # bad things happen if we close the nc file here.......
    return ret

  @reify
  def SounderCloudTemperature(self):
    import string, h2a
    from netCDF4 import Dataset
    from os.path import basename, isfile
    assert (len(self._hirs_files) == 1) # mjh TODO: generalize to n hirs files
    hirs = self._hirs_files[0] # TODO REDUNDNANT CODE
    cth_file = string.replace(basename(hirs), '.hdf', '')+'.ctp.nc' # mjh 3/18/2015 added .ctp
    if isfile(ctp_path+cth_file): # mjh 3/18/2015 fixed bug (path wasn't included)
        log.debug('cth_file found: %s' % cth_file)
    else:
        log.debug('cth_file NOT found, filling with junk: %s' % cth_file)
        return ((self.Latitude * 0.0) - 999.0).astype(np.float32)
    rootgrp = Dataset(ctp_path+cth_file, 'r')
    cth_data = np.array(rootgrp.variables['cloud_top_temperature'])
    ret = h2a.apply_mapping(self._HirsMapping, cth_data).astype(np.float32)
    ret[ret < 0.0] = -999.0
    #rootgrp.close()  # bad things happen if we close the nc file here.......
    return ret

  @reify
  def SounderCloudPressure(self):
    import string, h2a
    from netCDF4 import Dataset
    from os.path import basename, isfile
    assert (len(self._hirs_files) == 1) # mjh TODO: generalize to n hirs files
    hirs = self._hirs_files[0] # TODO REDUNDNANT CODE
    cth_file = string.replace(basename(hirs), '.hdf', '')+'.ctp.nc' # mjh 3/18/2015 added .ctp
    if isfile(ctp_path+cth_file): # mjh 3/18/2015 fixed bug (path wasn't included)
        log.debug('cth_file found: %s' % cth_file)
    else:
        log.debug('cth_file NOT found, filling with junk: %s' % cth_file)
        return ((self.Latitude * 0.0) - 999.0).astype(np.float32)
    rootgrp = Dataset(ctp_path+cth_file, 'r')
    cth_data = np.array(rootgrp.variables['cloud_top_pressure'])
    ret = h2a.apply_mapping(self._HirsMapping, cth_data).astype(np.float32)
    ret[ret < 0.0] = -999.0
    #rootgrp.close()  # bad things happen if we close the nc file here.......
    return ret

  @reify
  def HirsMask(self):
    elems = self.SounderElementIndex
    elems[elems != -999.0] = 1
    elems[elems == -999.0] = 0
    return elems.astype(np.int8)
