#!/usr/bin/env python
###################################
#
# Requires hirs_avhrr binary to work, set location in HIRS_AVHRR_BIN
#
###################################

import numpy as np
import numpy.ma as ma

from pyhdf.SD import *  #FIXME: don't import *

import subprocess

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

from config import *


def apply_mapping(mapping, data):
  datamapped = data[mapping[:,:,0], mapping[:,:,1]]
  datamasked = ma.array(datamapped, mask=mapping.mask[:,:,0], fill_value=-999.0)
  return datamasked.filled()


def hirs2avhrr_create_mapping(xTracks, yTracks, nfovs, imager_shape):
  # mjh:  hardcoded dim here breaks an assertion in avhrr.py (see HirsBands)
  mapping = np.empty((imager_shape[0],imager_shape[1],2), dtype=np.int)  # FIXME: should mask or use NaNs
  mapping.fill(-1)

  log.debug("mapping shape: %s" % str(mapping.shape))

  scanlines, scanangles = nfovs.shape
  
  for s in range(scanlines):
    for a in range(scanangles):
  
        nfov = nfovs[s][a]
        if nfov == 0:
          continue
  
        xs = xTracks[s][a]
        ys = yTracks[s][a]
  
        idxs = np.column_stack((ys, xs))
        for fov in range(nfov):
          yx = idxs[fov]
          y = yx[0] - 1
          x = yx[1] - 1

          mapping[y][x][0] = s
          mapping[y][x][1] = a

  mmapping = ma.masked_equal(mapping, -1)
  ma.set_fill_value(mmapping, 0)
  return mmapping


def hirs2avhrr_transform(sounder, imager):
  from os.path import basename
  # mjh 2014/08/21 D'OH!!! 24 processes trying to read/write to "collo.hdf" will be no good!!!
  collo_file = basename(sounder)[:-4]+"-collo.hdf"

  # FIXME: NOTE DIAMETER FOR HIRS_AVHRR - THIS IS SET FOR HIRS/4 ATM
  cmd = "%s %s %s 0.7 %s" % (HIRS_AVHRR_BIN, sounder, imager, collo_file)
  # mjh temporarily set to hirs/3 FOV for testing spot size characteristics
  #cmd = "%s %s %s 1.4 %s" % (HIRS_AVHRR_BIN, sounder, imager, collo_file)
  log.debug(cmd)
  subprocess.check_call(cmd, shell=True)
  
  # Grab the important bits out of the created file
  cv = SD(collo_file, SDC.READ)
  xTracks = cv.select('AVHRR_Across_Track_IDX')[:]   # max 3200
  yTracks = cv.select('AVHRR_Along_Track_IDX')[:]    # max 768
  nfovs = cv.select('Number_Of_AVHRR_FOV')[:]
  cv.end()

  imager_sd = SD(imager, SDC.READ)
  imager_shape = imager_sd.datasets()['longitude'][1]
  imager_sd.end()

  # Get rid of the cris_viirs file, we don't need it anymore
  cmd = "rm %s" % (collo_file)
  log.debug(cmd)
  subprocess.check_call(cmd, shell=True)
  
  log.debug("xTracks: %s yTracks: %s nfovs: %s" % (xTracks.shape, yTracks.shape, nfovs.shape))
  mapping = hirs2avhrr_create_mapping(xTracks, yTracks, nfovs, imager_shape)

  return mapping


